import json
import logging
import os

from octue.utils.decoders import OctueJSONDecoder
from octue.utils.encoders import OctueJSONEncoder


logger = logging.getLogger(__name__)


METADATA_FILENAME = ".octue"


def load_local_metadata_file(path=METADATA_FILENAME):
    """Load metadata from a local metadata records file, returning an empty dictionary if the file does not exist or is
    incorrectly formatted.

    :param str path:
    :return dict:
    """
    if not os.path.exists(path):
        return {}

    with open(path) as f:
        try:
            return json.load(f, cls=OctueJSONDecoder)
        except json.decoder.JSONDecodeError:
            logger.warning(
                f"The metadata file at {path!r} is incorrectly formatted so no metadata can be read from it. Please "
                "fix or delete it."
            )
            return {}


def overwrite_local_metadata_file(data, path=METADATA_FILENAME):
    """Create or overwrite the given local metadata file with the given data.

    :param dict data:
    :param str path:
    :return None:
    """
    with open(path, "w") as f:
        json.dump(data, f, cls=OctueJSONEncoder, indent=4)
        f.write("\n")



import json
import subprocess
import tempfile
import time
from abc import abstractmethod

import yaml

from octue.configuration import ServiceConfiguration
from octue.exceptions import DeploymentError


DOCKER_REGISTRY_URL = "eu.gcr.io"


class BaseDeployer:
    """An abstract tool for using an `octue.yaml` file in a repository to build and deploy the repository's `octue` app.
    This includes setting up a Google Cloud Build trigger, enabling automatic deployment during future development.
    Note that this tool requires the `gcloud` CLI to be available.

    This tool can be inherited from to create specific deployment tools (e.g. for Google Cloud Run or Google Dataflow)
    with the `deploy` and `_generate_cloud_build_configuration` methods overridden to set where and how the app is
    deployed. The `TOTAL_NUMBER_OF_STAGES` class variable should also be overridden and set to the number of stages in
    the subclass's deployment.

    The version information for the version of `gcloud` used to develop this tool is:
    ```
    Google Cloud SDK 367.0.0
    beta 2021.12.10
    bq 2.0.72
    cloud-build-local 0.5.2
    core 2021.12.10
    gsutil 5.5
    pubsub-emulator 0.6.0
    ```

    :param str octue_configuration_path: the path to the `octue.yaml` file if it's not in the current working directory
    :param str|None service_id: an ID to give the service if the one generated from `octue.yaml` isn't sufficient
    :return None:
    """

    TOTAL_NUMBER_OF_STAGES = None

    def __init__(self, octue_configuration_path, service_id=None, image_uri_template=None):
        self.service_configuration = ServiceConfiguration.from_file(octue_configuration_path)

        # Generated attributes.
        self.build_trigger_description = None
        self.generated_cloud_build_configuration = None
        self.service_id = (service_id or self.service_configuration.service_id).replace("/", ".")

        self.required_environment_variables = {"SERVICE_NAME": self.service_configuration.name}

        if service_id:
            self.required_environment_variables["SERVICE_ID"] = self.service_id

        self.image_uri_template = image_uri_template or (
            f"{DOCKER_REGISTRY_URL}/{self.service_configuration.project_name}/"
            f"{self.service_configuration.repository_name}/{self.service_configuration.name}:$SHORT_SHA"
        )

        self.success_message = f"[SUCCESS] Service deployed - it can be questioned via Pub/Sub at {service_id!r}."

    @abstractmethod
    def deploy(self, no_cache=False, update=False):
        """Deploy the octue app.

        :param bool no_cache: if `True`, don't use the Docker cache when building the image
        :param bool update: if `True`, allow the build trigger to already exist and just build and deploy a new image based on an updated `octue.yaml` file
        :return str: the service's UUID
        """

    @abstractmethod
    def _generate_cloud_build_configuration(self, no_cache=False):
        """Generate a Google Cloud Build configuration equivalent to a `cloudbuild.yaml` file in memory and assign it
        to the `generated_cloud_build_configuration` attribute.

        :param bool no_cache: if `True`, don't use the Docker cache when building the image
        :return None:
        """

    def _create_get_dockerfile_step(self, default_dockerfile_url):
        """If a Dockerfile step hasn't been provided, create a `cloudbuild.yaml` step that downloads the relevant
        default `octue` Dockerfile to build the image from. If it has been provided, provide an empty step.

        :param str default_dockerfile_url: the URL to get the default `octue` Dockerfile from
        :return (list, str): the `cloudbuild.yaml` step and the path to the Dockerfile
        """
        if self.service_configuration.dockerfile_path:
            return [], self.service_configuration.dockerfile_path

        # If no path to a dockerfile has been provided, add a step to download the default Octue service
        # Dockerfile to build the image from.
        get_dockerfile_step = [
            {
                "id": "Get default Octue Dockerfile",
                "name": "alpine:latest",
                "args": ["wget", default_dockerfile_url],
            }
        ]

        return get_dockerfile_step, "Dockerfile"

    def _create_build_secrets_sections(self):
        """Create the build secrets options for the Cloud Build configuration so secrets from the Google Cloud Secrets
        Manager can be used during the build step of the Cloud Build process. This provides the `availableSecrets`
        section for the overall configuration as well as Docker build args and the `secretEnv` section for the build
        step.

        :return (dict, dict): the `availableSecrets` section and the build secrets, the latter containing the Docker build args and `secretEnv` section for the Cloud Build build step
        """
        available_secrets_option = {}
        build_secrets = {"build_args": "", "secret_env": {}}

        if not self.service_configuration.secrets.get("build"):
            return available_secrets_option, build_secrets

        available_secrets_option = {
            "availableSecrets": {
                "secretManager": [
                    {
                        "versionName": f"projects/{self.service_configuration.project_name}/secrets/{secret_name}/versions/latest",
                        "env": secret_name,
                    }
                    for secret_name in self.service_configuration.secrets["build"]
                ]
            }
        }

        build_secrets["secret_env"] = {"secretEnv": self.service_configuration.secrets["build"]}

        build_secrets["build_args"] = [
            f"--build-arg={secret_name}=$${secret_name}" for secret_name in self.service_configuration.secrets["build"]
        ]

        return available_secrets_option, build_secrets

    def _create_build_trigger(self, update=False):
        """Create the build trigger in Google Cloud Build using the given `cloudbuild.yaml` file.

        :param bool update: if `True` and there is an existing trigger, delete it and replace it with an updated one
        :return None:
        """
        with ProgressMessage("Creating build trigger", 2, self.TOTAL_NUMBER_OF_STAGES) as progress_message:

            with tempfile.NamedTemporaryFile(delete=False) as temporary_file:

                if self.service_configuration.provided_cloud_build_configuration_path:
                    configuration_option = [
                        f"--build-config={self.service_configuration.provided_cloud_build_configuration_path}"
                    ]

                else:
                    # Put the Cloud Build configuration into a temporary file so it can be used by the `gcloud` command.
                    with open(temporary_file.name, "w") as f:
                        yaml.dump(self.generated_cloud_build_configuration, f)

                    configuration_option = [f"--inline-config={temporary_file.name}"]

                create_trigger_command = [
                    "gcloud",
                    f"--project={self.service_configuration.project_name}",
                    "beta",
                    "builds",
                    "triggers",
                    "create",
                    "github",
                    f"--name={self.service_configuration.name}",
                    f"--repo-name={self.service_configuration.repository_name}",
                    f"--repo-owner={self.service_configuration.repository_owner}",
                    f"--description={self.build_trigger_description}",
                    f"--branch-pattern={self.service_configuration.branch_pattern}",
                    *configuration_option,
                ]

                try:
                    self._run_command(create_trigger_command)
                except DeploymentError as e:
                    self._raise_or_ignore_already_exists_error(e, update, progress_message, finish_message="recreated.")

                    delete_trigger_command = [
                        "gcloud",
                        f"--project={self.service_configuration.project_name}",
                        "beta",
                        "builds",
                        "triggers",
                        "delete",
                        f"{self.service_configuration.name}",
                    ]

                    self._run_command(delete_trigger_command)
                    self._run_command(create_trigger_command)

    def _run_build_trigger(self):
        """Run the build trigger and return the build ID. The image URI is updated from the build metadata, ensuring
        that, if a `cloudbuild.yaml` file is provided instead of generated, the correct image URI from this file is
        used in later steps.

        :return None:
        """
        with ProgressMessage(
            f"Running build trigger on branch {self.service_configuration.branch_pattern!r}",
            3,
            self.TOTAL_NUMBER_OF_STAGES,
        ):

            build_command = [
                "gcloud",
                f"--project={self.service_configuration.project_name}",
                "--format=json",
                "beta",
                "builds",
                "triggers",
                "run",
                self.service_configuration.name,
                f"--branch={self.service_configuration.branch_pattern.strip('^$')}",
            ]

            process = self._run_command(build_command)
            metadata = json.loads(process.stdout.decode())["metadata"]
            self._wait_for_build_to_finish(metadata["build"]["id"])

    def _wait_for_build_to_finish(self, build_id, check_period=20):
        """Wait for the build with the given ID to finish.

        :param str build_id: the ID of the build to wait for
        :param float check_period: the period in seconds at which to check if the build has finished
        :return None:
        """
        get_build_command = [
            "gcloud",
            f"--project={self.service_configuration.project_name}",
            "--format=json",
            "builds",
            "describe",
            build_id,
        ]

        while True:
            process = self._run_command(get_build_command)
            status = json.loads(process.stdout.decode())["status"]

            if status not in {"QUEUED", "WORKING", "SUCCESS"}:
                raise DeploymentError(f"The build status is {status!r}.")

            if status == "SUCCESS":
                break

            time.sleep(check_period)

    @staticmethod
    def _get_short_head_commit_hash():
        """Get the short commit hash for the HEAD commit in the current git repository.

        :return str:
        """
        return subprocess.run(["git", "rev-parse", "--short", "HEAD"], capture_output=True).stdout.decode().strip()

    @staticmethod
    def _run_command(command):
        """Run a command in a subprocess, raising a `DeploymentError` if it fails.

        :param iter(str) command: the command to run in `subprocess` form e.g. `["cat", "my_file.txt"]`
        :raise octue.exceptions.DeploymentError: if the command fails
        :return subprocess.CompletedProcess:
        """
        process = subprocess.run(command, capture_output=True)

        if process.returncode != 0:
            raise DeploymentError(process.stderr.decode())

        return process

    @staticmethod
    def _raise_or_ignore_already_exists_error(exception, update, progress_message, finish_message="already exists."):
        """If `update` is `True` and the exception includes the words "already exists", ignore the exception and update
        the progress message's `finish_message`; otherwise, raise the exception.

        :param Exception exception: the exception to ignore or raise
        :param bool update: if `True`, ignore "already exists" errors but raise other errors
        :param ProgressMessage progress_message: the progress message to update with the finish message if appropriate
        :param str finish_message:
        :raise Exception:
        :return None:
        """
        if update and "already exists" in exception.args[0]:
            progress_message.finish_message = finish_message
            return

        raise exception


class ProgressMessage:
    """A context manager that, on entering the context, prints the given start message and, on leaving it, prints
    "done" on the same line. The use case is to surround a block of code with a start and finish message to give an
    idea of progress on the command line. If multiple progress messages are required, different instances of this class
    can be used and given information on their ordering and the total number of stages to produce an enumerated output.

    For example:
    ```
    [1/4] Generating Google Cloud Build configuration...done.
    [2/4] Creating build trigger...done.
    [3/4] Building and deploying service...done.
    [4/4] Creating Eventarc Pub/Sub run trigger...done.
    ```

    :param str start_message: the message to print before the code in the context is executed
    :param int stage: the position of the progress message among all the related progress messages
    :param int total_number_of_stages: the total number of progress messages that will be printed
    :return None:
    """

    def __init__(self, start_message, stage, total_number_of_stages):
        self.start_message = f"[{stage}/{total_number_of_stages}] {start_message}..."
        self.finish_message = "done."

    def __enter__(self):
        """Print the start message.

        :return ProgressMessage:
        """
        print(self.start_message, end="", flush=True)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Print the finish message on the same line as the start message. If there's an error, print "ERROR" instead.

        :return None:
        """
        if exc_type:
            print("ERROR.")
        else:
            print(self.finish_message)



from dateparser import parse


def read_dat_file(file):
    """Read a dat file containing meteorological mast metadata

    You don't really need to care about this, because your files are unlikely to be in the same form as our
    example dat files. You'll need to build your own file readers and test them (or use open-source libraries to do
    the reading).

    So this is only here to make the example work. Although you could use this as a case study of how annoying it is to
    parse custom formatted data files - even "simple" ones!

    :parameter file: File from the dataset that contains metadata
    :type file: octue.Datafile

    :return: Dictionary of available metadata
    :rtype: dict
    """
    # Read the file
    with open(file.local_path, "r") as text_file:
        lines = text_file.readlines()

    # Parse the metadata
    metadata = dict()
    for line in lines:
        # In the proprietary example dat file, tabs separate parameters from their value, so if a line contains a tab,
        # import that parameter-value pair
        if ":" in line:
            param = line.split(":", 1)
            key = param[0].strip().lower()
            value = param[1].strip()
            value = parse(value) if key == "date" else float(value)
            metadata[key] = value

    # Make sure that a `date` field is in the metadata, as we'll need it later
    if "date" not in metadata.keys():
        raise ValueError("No DATE field in the metadata file (required)")

    return metadata



import importlib
import logging
import os
import re
import sys

import google.api_core.exceptions
from google import auth
from google.cloud import secretmanager
from jsonschema import ValidationError, validate as jsonschema_validate

import twined.exceptions
from octue import exceptions
from octue.log_handlers import apply_log_handler, create_octue_formatter, get_log_record_attributes_for_environment
from octue.resources import Child
from octue.resources.analysis import CLASS_MAP, Analysis
from octue.utils import gen_uuid
from twined import Twine


logger = logging.getLogger(__name__)


class Runner:
    """A runner of analyses for a given service.

    The Runner class provides a set of configuration parameters for use by your application, together with a range of
    methods for managing input and output file parsing as well as controlling logging.

    :param Union[AppFrom, callable, str] app_src: Either an instance of the AppFrom manager class which has a run() method, or a function which accepts a single parameter (the instantiated analysis), or a string pointing to an application folder (which should contain an 'app.py' function like the templates)
    :param str|twined.Twine twine: path to the twine file, a string containing valid twine json, or a Twine instance
    :param str|dict|_io.TextIOWrapper|None configuration_values: The strand data. Can be expressed as a string path of a *.json file (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an already-parsed dict.
    :param str|dict|_io.TextIOWrapper|None configuration_manifest: The strand data. Can be expressed as a string path of a *.json file (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an already-parsed dict.
    :param Union[str, dict, None] children: The children strand data. Can be expressed as a string path of a *.json file (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an already-parsed dict.
    :param str|None output_location: the path to a cloud directory to save output datasets at
    :param str|None project_name: name of Google Cloud project to get credentials from
    :param str|None service_id: the ID of the service being run
    :return None:
    """

    def __init__(
        self,
        app_src,
        twine="twine.json",
        configuration_values=None,
        configuration_manifest=None,
        children=None,
        output_location=None,
        project_name=None,
        service_id=None,
    ):
        self.app_src = app_src
        self.children = children

        if output_location and not re.match(r"^gs://[a-z\d][a-z\d_./-]*$", output_location):
            raise exceptions.InvalidInputException(
                "The output location must be a Google Cloud Storage path e.g. 'gs://bucket-name/output_directory'."
            )

        self.output_location = output_location

        # Ensure the twine is present and instantiate it.
        if isinstance(twine, Twine):
            self.twine = twine
        else:
            self.twine = Twine(source=twine)

        logger.debug("Parsed twine with strands %r", self.twine.available_strands)

        # Validate and initialise configuration data.
        self.configuration = self.twine.validate(
            configuration_values=configuration_values,
            configuration_manifest=configuration_manifest,
            cls=CLASS_MAP,
        )
        logger.debug("Configuration validated.")

        self.service_id = service_id
        self._project_name = project_name

    def run(
        self,
        analysis_id=None,
        input_values=None,
        input_manifest=None,
        analysis_log_level=logging.INFO,
        analysis_log_handler=None,
        handle_monitor_message=None,
    ):
        """Run an analysis.

        :param str|None analysis_id: UUID of analysis
        :param Union[str, dict, None] input_values: the input_values strand data. Can be expressed as a string path of a *.json file (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an already-parsed dict.
        :param Union[str, octue.resources.manifest.Manifest, None] input_manifest: The input_manifest strand data. Can be expressed as a string path of a *.json file (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an already-parsed dict.
        :param str analysis_log_level: the level below which to ignore log messages
        :param logging.Handler|None analysis_log_handler: the logging.Handler instance which will be used to handle logs for this analysis run. Handlers can be created as per the logging cookbook https://docs.python.org/3/howto/logging-cookbook.html but should use the format defined above in LOG_FORMAT.
        :param callable|None handle_monitor_message: a function that sends monitor messages to the parent that requested the analysis
        :return None:
        """
        if hasattr(self.twine, "credentials"):
            self._populate_environment_with_google_cloud_secrets()
            credentials = self.twine.credentials
        else:
            credentials = None

        inputs = self.twine.validate(
            input_values=input_values,
            input_manifest=input_manifest,
            credentials=credentials,
            children=self.children,
            cls=CLASS_MAP,
            allow_missing=False,
            allow_extra=False,
        )
        logger.debug("Inputs validated.")

        for manifest_strand in self.twine.available_manifest_strands:
            if manifest_strand == "output_manifest":
                continue

            self._validate_dataset_file_tags(manifest_kind=manifest_strand, manifest=inputs[manifest_strand])

        if inputs["children"] is not None:
            inputs["children"] = {
                child["key"]: Child(
                    name=child["key"],
                    id=child["id"],
                    backend=child["backend"],
                    internal_service_name=self.service_id,
                )
                for child in inputs["children"]
            }

        outputs_and_monitors = self.twine.prepare("monitor_message", "output_values", "output_manifest", cls=CLASS_MAP)

        analysis_id = str(analysis_id) if analysis_id else gen_uuid()

        if analysis_log_handler:
            extra_log_handlers = [analysis_log_handler]
        else:
            extra_log_handlers = None

        # Temporarily replace the root logger's handlers with a `StreamHandler` and the analysis log handler that
        # include the analysis ID in the logging metadata.
        with AnalysisLogHandlerSwitcher(
            analysis_id=analysis_id,
            logger=logging.getLogger(),
            analysis_log_level=analysis_log_level,
            extra_log_handlers=extra_log_handlers,
        ):

            analysis = Analysis(
                id=analysis_id,
                twine=self.twine,
                handle_monitor_message=handle_monitor_message,
                output_location=self.output_location,
                **self.configuration,
                **inputs,
                **outputs_and_monitors,
            )

            try:
                # App as a class that takes "analysis" as a constructor argument and contains a method named "run" that
                # takes no arguments.
                if isinstance(self.app_src, type):
                    self.app_src(analysis).run()

                # App as a module containing a function named "run" that takes "analysis" as an argument.
                elif hasattr(self.app_src, "run"):
                    self.app_src.run(analysis)

                # App as a string path to a module containing a class named "App" or a function named "run". The same other
                # specifications apply as described above.
                elif isinstance(self.app_src, str):

                    with AppFrom(self.app_src) as app:
                        if hasattr(app.app_module, "App"):
                            app.app_module.App(analysis).run()
                        else:
                            app.run(analysis)

                # App as a function that takes "analysis" as an argument.
                else:
                    self.app_src(analysis)

            except ModuleNotFoundError as e:
                raise ModuleNotFoundError(f"{e.msg} in {os.path.abspath(self.app_src)!r}.")

            except Exception as e:
                logger.error(str(e))
                raise e

            return analysis

    def _populate_environment_with_google_cloud_secrets(self):
        """Get any secrets specified in the credentials strand from Google Cloud Secret Manager and put them in the
        local environment, ready for use by the runner.

        :return None:
        """
        missing_credentials = tuple(
            credential for credential in self.twine.credentials if credential["name"] not in os.environ
        )

        if not missing_credentials:
            return

        google_cloud_credentials, project_name = auth.default()
        secrets_client = secretmanager.SecretManagerServiceClient(credentials=google_cloud_credentials)

        if google_cloud_credentials is None:
            project_name = self._project_name

        for credential in missing_credentials:
            secret_path = secrets_client.secret_version_path(
                project=project_name, secret=credential["name"], secret_version="latest"
            )

            try:
                secret = secrets_client.access_secret_version(name=secret_path).payload.data.decode("UTF-8")
            except google.api_core.exceptions.NotFound:
                # No need to raise an error here as the Twine validation that follows will do so.
                continue

            os.environ[credential["name"]] = secret

    def _validate_dataset_file_tags(self, manifest_kind, manifest):
        """Validate the tags of the files of each dataset in the manifest against the file tags template in the
        corresponding dataset field in the given manifest field of the twine.

        :param str manifest_kind: the kind of manifest that's being validated (so the correct schema can be accessed)
        :param octue.resources.manifest.Manifest manifest: the manifest whose datasets' files are to be validated
        :return None:
        """
        # This is the manifest schema included in the twine.json file, not the schema for `manifest.json` files.
        manifest_schema = getattr(self.twine, manifest_kind)

        for dataset_name, dataset_schema in manifest_schema["datasets"].items():
            dataset = manifest.datasets.get(dataset_name)
            file_tags_template = dataset_schema.get("file_tags_template")

            # Allow optional datasets in future (not currently allowed by `twined`).
            if not (dataset and file_tags_template):
                continue

            for file in dataset.files:
                try:
                    jsonschema_validate(instance=dict(file.tags), schema=file_tags_template)
                except ValidationError as e:
                    raise twined.exceptions.invalid_contents_map[manifest_kind](str(e))


def unwrap(fcn):
    """Recurse through wrapping to get the raw function without decorators."""
    if hasattr(fcn, "__wrapped__"):
        return unwrap(fcn.__wrapped__)
    return fcn


class AppFrom:
    """A context manager that imports the module "app" from a file named "app.py" in the given directory on entry (by
    making a temporary addition to the system path) and unloads it (by deleting it from `sys.modules`) on exit. It will
    issue a warning if an existing module called "app" is already loaded. Usage example:

    ```python3
    with AppFrom('/path/to/dir') as app:
        Runner().run(app)
    ```

    :param str app_path: path to directory containing module named "app.py".
    :return None:
    """

    def __init__(self, app_path="."):
        self.app_path = os.path.abspath(os.path.normpath(app_path))
        logger.debug("Initialising AppFrom context at app_path %s", self.app_path)
        self.app_module = None

    def __enter__(self):
        # Warn on an app present on the system path
        if "app" in sys.modules.keys():
            logger.warning(
                "Module 'app' already on system path. Using 'AppFrom' context will yield unexpected results. Avoid "
                "using 'app' as a python module, except for your main entrypoint."
            )

        # Insert the present directory first on the system path.
        sys.path.insert(0, self.app_path)

        # Import the app from the present directory.
        self.app_module = importlib.import_module("app")

        # Immediately clean up the entry to the system path (don't use "remove" because if the user has it in their
        # path, this'll be an unexpected side effect, and don't do it in cleanup in case the called code inserts a path)
        sys.path.pop(0)
        logger.debug("Imported app at app_path and cleaned up temporary modification to sys.path %s", self.app_path)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Unload the imported module.

        :return None:
        """
        try:
            del sys.modules["app"]
            logger.debug("Deleted app from sys.modules")

        except KeyError:
            context_manager_name = type(self).__name__
            logger.warning(
                f"Module 'app' was already removed from the system path prior to exiting the {context_manager_name} "
                f"context manager. Using the {context_manager_name} context may yield unexpected results."
            )

    @property
    def run(self):
        """Get the unwrapped run function from app.py in the application's root directory."""
        return unwrap(self.app_module.run)


class AnalysisLogHandlerSwitcher:
    """A context manager that, when activated, takes the given logger, removes its handlers, and adds a local handler
    and any other handlers provided to it. A formatter is applied to the handlers that includes the given analysis ID
    in the logging context. On leaving the context, the logger's initial handlers are restored to it and any that were
    added to it in the context are removed.

    :param str analysis_id:
    :param logger.Logger logger:
    :param str analysis_log_level:
    :param list(logging.Handler) extra_log_handlers:
    :return None:
    """

    def __init__(self, analysis_id, logger, analysis_log_level, extra_log_handlers=None):
        self.analysis_id = analysis_id
        self.logger = logger
        self.analysis_log_level = analysis_log_level
        self.extra_log_handlers = extra_log_handlers or []
        self.initial_handlers = []

    def __enter__(self):
        """Remove the initial handlers from the logger, create a formatter that includes the analysis ID, use the
        formatter on the local and extra handlers, and add these handlers to the logger.

        :return None:
        """
        self.initial_handlers = list(self.logger.handlers)
        self._remove_log_handlers()

        # Add the analysis ID to the logging metadata.
        formatter = create_octue_formatter(
            get_log_record_attributes_for_environment(),
            [f"analysis-{self.analysis_id}"],
        )

        # Apply a local console `StreamHandler` to the logger.
        apply_log_handler(formatter=formatter, log_level=self.analysis_log_level)

        if not self.extra_log_handlers:
            return

        # Apply any other given handlers to the logger.
        for extra_handler in self.extra_log_handlers:
            apply_log_handler(handler=extra_handler, log_level=self.analysis_log_level, formatter=formatter)

    def __exit__(self, *args):
        """Remove the new handlers from the logger and re-add the initial handlers.

        :return None:
        """
        self._remove_log_handlers()

        for handler in self.initial_handlers:
            self.logger.addHandler(handler)

    def _remove_log_handlers(self):
        """Remove all handlers from the logger.

        :return None:
        """
        for handler in list(self.logger.handlers):
            self.logger.removeHandler(handler)



import logging

from flask import Flask, request

from octue.cloud.deployment.google.answer_pub_sub_question import answer_question
from octue.utils.metadata import load_local_metadata_file, overwrite_local_metadata_file


logger = logging.getLogger(__name__)
app = Flask(__name__)


@app.route("/", methods=["POST"])
def index():
    """Receive questions from Google Cloud Run in the form of Google Pub/Sub messages.

    :return (str, int):
    """
    envelope = request.get_json()

    if not envelope:
        return _log_bad_request_and_return_400_response("No Pub/Sub message received.")

    if not isinstance(envelope, dict) or "message" not in envelope:
        return _log_bad_request_and_return_400_response("Invalid Pub/Sub message format.")

    question = envelope["message"]

    if "data" not in question or "attributes" not in question or "question_uuid" not in question["attributes"]:
        return _log_bad_request_and_return_400_response("Invalid Pub/Sub message format.")

    question_uuid = question["attributes"]["question_uuid"]

    local_metadata = load_local_metadata_file()
    delivered_questions = local_metadata.get("delivered_questions", set())

    # Acknowledge questions that are redelivered to stop further redelivery and redundant processing.
    if question_uuid in delivered_questions:
        logger.info(
            "Question %r has already been received by the service. It will now be acknowledged to prevent further "
            "redundant redelivery.",
            question_uuid,
        )
        return ("", 204)

    # Otherwise add the question UUID to the set.
    delivered_questions.add(question_uuid)
    overwrite_local_metadata_file(local_metadata)

    project_name = envelope["subscription"].split("/")[1]
    answer_question(question=question, project_name=project_name)
    return ("", 204)


def _log_bad_request_and_return_400_response(message):
    """Log an error return a bad request (400) response.

    :param str message:
    :return (str, int):
    """
    logger.error(message)
    return (f"Bad Request: {message}", 400)



class OctueSDKException(Exception):
    """All exceptions from this library inherit from here"""


class InvalidInputException(OctueSDKException, ValueError):
    """Raise when an object is instantiated or a function called with invalid inputs"""


class FileNotFoundException(InvalidInputException, FileNotFoundError):
    """Raise when a required folder (e.g. <data_dir>/input) cannot be found"""


class FileLocationError(Exception):
    """Raise when a file exists in an unsupported location for a given operation."""


class ProtectedAttributeException(OctueSDKException, KeyError):
    """Raise when a user attempts to set an attribute whose value should be protected"""


class FolderNotFoundException(InvalidInputException):
    """Raise when a multi manifest can not be refined to a single manifest in a search"""


class ManifestNotFoundException(InvalidInputException):
    """Raise when a multi manifest can not be refined to a single manifest in a search"""


class InvalidManifestException(InvalidInputException):
    """Raise when a manifest loaded from JSON does not pass validation"""


class InvalidManifestTypeException(InvalidManifestException):
    """Raised when user attempts to create a manifest of a type other than 'input', 'output' or 'build'"""


class InvalidOctueFileTypeException(OctueSDKException):
    """Raised when you attempt to register a file type in the results manifest that Octue doesn't know about"""


class InvalidFilePointerException(OctueSDKException):
    """Raised when you attempt to create a Datafile with an object that is not file-like."""


class NotImplementedYetException(OctueSDKException):
    """Raised when you attempt to use a function whose high-level API is in place, but which is not implemented yet"""


class UnexpectedNumberOfResultsException(OctueSDKException):
    """Raise when searching for a single data file (or a particular number of data files) and the number of results
    exceeds that expected
    """


class BrokenSequenceException(OctueSDKException):
    """Raise when filtering for a sequence of files whose numbering does not start at 0, or does not monotonically
    increase
    """


class InvalidTagException(OctueSDKException, ValueError):
    """Raise when a tag applied to a data file or dataset"""


class InvalidLabelException(OctueSDKException, ValueError):
    """Raise when a label applied to a data file or dataset"""


class ServiceNotFound(OctueSDKException):
    """Raise when a Service of the given ID has not been found on the Google Pub/Sub server (i.e. if there is no topic
    associated with the Service ID).
    """


class QuestionNotDelivered(OctueSDKException):
    """Raise if a question has not been delivered or no acknowledgement of its delivery has been received."""


class BackendNotFound(OctueSDKException):
    """Raise when details of a backend that doesn't exist in `octue.resources.service_backends` are given for use as a
    Service backend.
    """


class AttributeConflict(OctueSDKException):
    """Raise if, when trying to set an attribute whose current value has a significantly higher confidence than the new
    value, the new value conflicts with the current value.
    """


class MissingServiceID(OctueSDKException):
    """Raise when a specific ID for a service is expected to be provided, but is missing or None."""


class CloudLocationNotSpecified(OctueSDKException):
    """Raise when attempting to interact with a cloud resource implicitly but the implicit details of its location are
    missing.
    """


class InvalidMonitorMessage(OctueSDKException):
    """Raise if a monitor message fails validation against the "monitor_message_schema" field of the Twine."""


class DeploymentError(OctueSDKException):
    """Raise if an error occurs during the deployment of an octue service."""


class CloudStorageBucketNotFound(OctueSDKException):
    """Raise if attempting to access a cloud storage bucket that cannot be found."""


class PushSubscriptionCannotBePulled(OctueSDKException):
    """Raise if attempting to pull a push subscription."""


class ReadOnlyResource(OctueSDKException):
    """Raise if attempting to alter a read-only resource."""



import json
import logging
import os
import re
import time

from google.api_core import retry

import octue.exceptions
import twined.exceptions
from octue.log_handlers import COLOUR_PALETTE
from octue.resources.manifest import Manifest
from octue.utils.exceptions import create_exceptions_mapping


if os.environ.get("COMPUTE_PROVIDER", "UNKNOWN") in {"GOOGLE_CLOUD_RUN", "GOOGLE_DATAFLOW"}:
    # Google Cloud logs don't support colour currently - provide a no-operation function.
    colourise = lambda string, text_colour=None, background_colour=None: string
else:
    from octue.utils.colour import colourise


logger = logging.getLogger(__name__)

EXCEPTIONS_MAPPING = create_exceptions_mapping(
    globals()["__builtins__"], vars(twined.exceptions), vars(octue.exceptions)
)


class OrderedMessageHandler:
    """A handler for Google Pub/Sub messages that ensures messages are handled in the order they were sent.

    :param google.pubsub_v1.services.subscriber.client.SubscriberClient subscriber: a Google Pub/Sub subscriber
    :param octue.cloud.pub_sub.subscription.Subscription subscription: the subscription messages are pulled from
    :param callable|None handle_monitor_message: a function to handle monitor messages (e.g. send them to an endpoint for plotting or displaying) - this function should take a single JSON-compatible python primitive
    :param str service_name: an arbitrary name to refer to the service subscribed to by (used for labelling its remote log messages)
    :param dict|None message_handlers: a mapping of message handler names to callables that handle each type of message
    :return None:
    """

    def __init__(
        self,
        subscriber,
        subscription,
        handle_monitor_message=None,
        service_name="REMOTE",
        message_handlers=None,
    ):
        self.subscriber = subscriber
        self.subscription = subscription
        self.handle_monitor_message = handle_monitor_message
        self.service_name = service_name

        self.received_delivery_acknowledgement = None
        self._start_time = time.perf_counter()
        self._waiting_messages = None
        self._previous_message_number = -1

        self._message_handlers = message_handlers or {
            "delivery_acknowledgement": self._handle_delivery_acknowledgement,
            "monitor_message": self._handle_monitor_message,
            "log_record": self._handle_log_message,
            "exception": self._handle_exception,
            "result": self._handle_result,
        }

        self._log_message_colours = [COLOUR_PALETTE[1], *COLOUR_PALETTE[3:]]

    def handle_messages(self, timeout=60, delivery_acknowledgement_timeout=120):
        """Pull messages and handle them in the order they were sent until a result is returned by a message handler,
        then return that result.

        :param float|None timeout: how long to wait for an answer before raising a `TimeoutError`
        :param float delivery_acknowledgement_timeout: how long to wait for a delivery acknowledgement before raising `QuestionNotDelivered`
        :raise TimeoutError: if the timeout is exceeded before receiving the final message
        :return dict:
        """
        self.received_delivery_acknowledgement = False
        self._waiting_messages = {}
        self._previous_message_number = -1

        pull_timeout = None

        while True:

            if timeout is not None:
                run_time = time.perf_counter() - self._start_time

                if run_time > timeout:
                    raise TimeoutError(
                        f"No final answer received from topic {self.subscription.topic.path!r} after {timeout} seconds.",
                    )

                pull_timeout = timeout - run_time

            message = self._pull_message(
                timeout=pull_timeout,
                delivery_acknowledgement_timeout=delivery_acknowledgement_timeout,
            )

            self._waiting_messages[int(message["message_number"])] = message

            try:
                while self._waiting_messages:
                    message = self._waiting_messages.pop(self._previous_message_number + 1)
                    result = self._handle_message(message)

                    if result is not None:
                        return result

            except KeyError:
                pass

    def _pull_message(self, timeout, delivery_acknowledgement_timeout):
        """Pull a message from the subscription, raising a `TimeoutError` if the timeout is exceeded before succeeding.

        :param float|None timeout: how long to wait in seconds for the message before raising a `TimeoutError`
        :param float delivery_acknowledgement_timeout: how long to wait for a delivery acknowledgement before raising `QuestionNotDelivered`
        :raise TimeoutError|concurrent.futures.TimeoutError: if the timeout is exceeded
        :raise octue.exceptions.QuestionNotDelivered: if a delivery acknowledgement is not received in time
        :return dict: message containing data
        """
        start_time = time.perf_counter()

        while True:
            attempt = 1

            while True:
                logger.debug("Pulling messages from Google Pub/Sub: attempt %d.", attempt)

                pull_response = self.subscriber.pull(
                    request={"subscription": self.subscription.path, "max_messages": 1},
                    retry=retry.Retry(),
                )

                try:
                    answer = pull_response.received_messages[0]
                    break

                except IndexError:
                    logger.debug("Google Pub/Sub pull response timed out early.")
                    attempt += 1

                    run_time = time.perf_counter() - start_time

                    if timeout is not None and run_time > timeout:
                        raise TimeoutError(
                            f"No message received from topic {self.subscription.topic.path!r} after {timeout} seconds.",
                        )

                    if not self.received_delivery_acknowledgement:
                        if run_time > delivery_acknowledgement_timeout:
                            raise octue.exceptions.QuestionNotDelivered(
                                f"No delivery acknowledgement received for topic {self.subscription.topic.path!r} "
                                f"after {delivery_acknowledgement_timeout} seconds."
                            )

            self.subscriber.acknowledge(request={"subscription": self.subscription.path, "ack_ids": [answer.ack_id]})

            logger.debug(
                "%r received a message related to question %r.",
                self.subscription.topic.service,
                self.subscription.topic.path.split(".")[-1],
            )

            return json.loads(answer.message.data.decode())

    def _handle_message(self, message):
        """Pass a message to its handler and update the previous message number.

        :param dict message:
        :return dict|None:
        """
        self._previous_message_number += 1

        try:
            return self._message_handlers[message["type"]](message)
        except KeyError:
            logger.warning("Received a message of unknown type %r.", message["type"])

    def _handle_delivery_acknowledgement(self, message):
        """Mark the question as delivered to prevent resending it.

        :param dict message:
        :return None:
        """
        self.received_delivery_acknowledgement = True
        logger.info("%r's question was delivered at %s.", self.subscription.topic.service, message["delivery_time"])

    def _handle_monitor_message(self, message):
        """Send a monitor message to the handler if one has been provided.

        :param dict message:
        :return None:
        """
        logger.debug("%r received a monitor message.", self.subscription.topic.service)

        if self.handle_monitor_message is not None:
            self.handle_monitor_message(json.loads(message["data"]))

    def _handle_log_message(self, message):
        """Deserialise the message into a log record and pass it to the local log handlers, adding [<service-name>] to
        the start of the log message.

        :param dict message:
        :return None:
        """
        record = logging.makeLogRecord(message["log_record"])

        # Add information about the immediate child sending the message and colour it with the first colour in the
        # colour palette.
        immediate_child_analysis_section = colourise(
            f"[{self.service_name} | analysis-{message['analysis_id']}]",
            text_colour=self._log_message_colours[0],
        )

        # Colour any analysis sections from children of the immediate child with the rest of the colour palette and
        # colour the message from the furthest child white.
        subchild_analysis_sections = [section.strip("[") for section in re.split("] ", record.msg)]
        final_message = colourise(subchild_analysis_sections.pop(-1))

        for i in range(len(subchild_analysis_sections)):
            subchild_analysis_sections[i] = colourise(
                "[" + subchild_analysis_sections[i] + "]",
                text_colour=self._log_message_colours[1:][i % len(self._log_message_colours[1:])],
            )

        record.msg = " ".join([immediate_child_analysis_section, *subchild_analysis_sections, final_message])
        logger.handle(record)

    def _handle_exception(self, message):
        """Raise the exception from the responding service that is serialised in `data`.

        :param dict message:
        :raise Exception:
        :return None:
        """
        exception_message = "\n\n".join(
            (
                message["exception_message"],
                f"The following traceback was captured from the remote service {self.service_name!r}:",
                "".join(message["traceback"]),
            )
        )

        try:
            raise EXCEPTIONS_MAPPING[message["exception_type"]](exception_message)

        # Allow unknown exception types to still be raised.
        except KeyError:
            raise type(message["exception_type"], (Exception,), {})(exception_message)

    def _handle_result(self, message):
        """Convert the result to the correct form, deserialising the output manifest if it is present in the message.

        :param dict message:
        :return dict:
        """
        logger.info("Received an answer to question %r.", self.subscription.topic.path.split(".")[-1])

        if message["output_manifest"] is None:
            output_manifest = None
        else:
            output_manifest = Manifest.deserialise(message["output_manifest"], from_string=True)

        return {"output_values": message["output_values"], "output_manifest": output_manifest}



from abc import abstractmethod

import pkg_resources


class Metadata:
    _METADATA_ATTRIBUTES = tuple()

    def metadata(self, include_sdk_version=True):
        """Get the instance's metadata in primitive form.

        :param bool include_sdk_version: if `True`, include the `octue` version that instantiated the instance in the metadata
        :return dict:
        """
        metadata = {name: getattr(self, name) for name in self._METADATA_ATTRIBUTES}

        if include_sdk_version:
            metadata["sdk_version"] = pkg_resources.get_distribution("octue").version

        return metadata

    @abstractmethod
    def _set_metadata(self, metadata):
        """Set the instance's metadata.

        :param dict metadata:
        :return None:
        """



import copy
import datetime
import functools
import json
import logging
import os
import shutil
import tempfile

import google.api_core.exceptions
import pkg_resources
import requests
from google_crc32c import Checksum

from octue.resources.label import LabelSet
from octue.resources.tag import TagDict


try:
    import h5py
except ModuleNotFoundError:
    pass

from octue.cloud import storage
from octue.cloud.storage import GoogleCloudStorageClient
from octue.exceptions import CloudLocationNotSpecified, ReadOnlyResource
from octue.mixins import CloudPathable, Filterable, Hashable, Identifiable, Labelable, Metadata, Serialisable, Taggable
from octue.mixins.hashable import EMPTY_STRING_HASH_VALUE
from octue.utils.decoders import OctueJSONDecoder
from octue.utils.metadata import METADATA_FILENAME, load_local_metadata_file, overwrite_local_metadata_file


logger = logging.getLogger(__name__)


OCTUE_METADATA_NAMESPACE = "octue"


class Datafile(Labelable, Taggable, Serialisable, Identifiable, Hashable, Filterable, Metadata, CloudPathable):
    """A representation of a data file on the Octue system. Metadata for the file is obtained from its corresponding
    cloud object or a local `.octue` metadata file, if present. If no stored metadata is available, it can be set during
    or after instantiation.

    :param str|None path: The path of this file locally or in the cloud, which may include folders or subfolders, within the dataset
    :param str|None local_path: If a cloud path is given as the `path` parameter, this is the path to an existing local file that is known to be in sync with the cloud object
    :param str|None cloud_path: If a local path is given for the `path` parameter, this is a cloud path to keep in sync with the local file
    :param datetime.datetime|int|float|None timestamp: A posix timestamp associated with the file, in seconds since epoch, typically when it was created but could relate to a relevant time point for the data
    :param str mode: if using as a context manager, open the datafile for reading/editing in this mode (the mode options are the same as for the builtin `open` function)
    :param bool update_metadata: if using as a context manager and this is `True`, update the stored metadata of the datafile when the context is exited
    :param bool hypothetical: if `True`, ignore any metadata stored for this datafile locally or in the cloud and use whatever is given at instantiation
    :param str id: The Universally Unique ID of this file (checked to be valid if not None, generated if None)
    :param dict|octue.resources.tag.TagDict|None tags: key-value pairs with string keys conforming to the Octue tag format (see `TagDict`)
    :param iter(str)|octue.resources.label.LabelSet|None labels: Space-separated string of labels relevant to this file
    :return None:
    """

    _CLOUD_PATH_ATTRIBUTE_NAME = "cloud_path"

    _METADATA_ATTRIBUTES = ("id", "timestamp", "tags", "labels")

    _SERIALISE_FIELDS = (
        *_METADATA_ATTRIBUTES,
        "name",
        "path",
        "cloud_path",
        "_cloud_metadata",
    )

    def __init__(
        self,
        path,
        local_path=None,
        cloud_path=None,
        timestamp=None,
        mode="r",
        update_metadata=True,
        hypothetical=False,
        id=None,
        tags=None,
        labels=None,
        **kwargs,
    ):
        super().__init__(
            id=id,
            name=kwargs.pop("name", None),
            immutable_hash_value=kwargs.pop("immutable_hash_value", None),
            tags=tags,
            labels=labels,
        )

        self.timestamp = timestamp
        self._open_attributes = {"mode": mode, "update_metadata": update_metadata, **kwargs}
        self._local_path = None
        self._cloud_path = None
        self._cloud_metadata = {}

        if storage.path.is_cloud_path(path):
            self._instantiate_from_cloud_object(path, local_path, ignore_stored_metadata=hypothetical)
        else:
            self._instantiate_from_local_path(path, cloud_path, ignore_stored_metadata=hypothetical)

        if hypothetical:
            logger.debug("Ignored stored metadata for %r.", self)
        else:
            if self.metadata(use_octue_namespace=False, include_sdk_version=False) != {
                "id": id or self.id,
                "timestamp": timestamp,
                "tags": TagDict(tags),
                "labels": LabelSet(labels),
            }:
                logger.warning(
                    "Overriding metadata given at instantiation with stored metadata for %r - set `hypothetical` to "
                    "`True` at instantiation to avoid this.",
                    self,
                )

    @classmethod
    def deserialise(cls, serialised_datafile, from_string=False):
        """Deserialise a Datafile from a dictionary or JSON string.

        :param dict|str serialised_datafile:
        :param bool from_string:
        :return Datafile:
        """
        if from_string:
            serialised_datafile = json.loads(serialised_datafile, cls=OctueJSONDecoder)
        else:
            serialised_datafile = copy.deepcopy(serialised_datafile)

        cloud_metadata = serialised_datafile.pop("_cloud_metadata", {})
        datafile = Datafile(**serialised_datafile)
        datafile._cloud_metadata = cloud_metadata
        return datafile

    @property
    def name(self):
        """Get the name of the datafile.

        :return str:
        """
        if self._name:
            return self._name

        if self.exists_in_cloud:
            return str(storage.path.split(self.cloud_path)[-1]).split("?")[0]

        return str(os.path.split(self.local_path)[-1])

    @property
    def extension(self):
        """Get the extension of the datafile.

        :return str:
        """
        return os.path.splitext(self.name)[-1].strip(".")

    @property
    def cloud_path(self):
        """Get the cloud path of the datafile.

        :return str|None:
        """
        return self._cloud_path

    @cloud_path.setter
    def cloud_path(self, path):
        """Set the cloud path of the datafile.

        :param str|None path:
        :return None:
        """
        if path is None:

            if not self.exists_locally:
                raise CloudLocationNotSpecified(
                    "The cloud path cannot be reset because this datafile only exists in the cloud."
                )

            self._cloud_path = None

        else:
            self.upload(cloud_path=path)

    @property
    def cloud_hash_value(self):
        """Get the hash value of the datafile according to its cloud file.

        :return str|None: `None` if no cloud metadata is available
        """
        return self._cloud_metadata.get("crc32c")

    @property
    def timestamp(self):
        """Get the timestamp of the datafile.

        :return float:
        """
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value):
        """Set the datafile's timestamp.

        :param datetime.datetime|int|float|None value:
        :raise TypeError: if value is of an incorrect type
        :return None:
        """
        if isinstance(value, datetime.datetime) or value is None:
            self._timestamp = value
        elif isinstance(value, (int, float)):
            self._timestamp = datetime.datetime.fromtimestamp(value)
        else:
            raise TypeError(
                f"timestamp should be a datetime.datetime instance, an int, a float, or None; received {value!r}"
            )

    @property
    def posix_timestamp(self):
        """Get the timestamp of the datafile in posix format.

        :return float:
        """
        if self.timestamp is None:
            return None

        return self.timestamp.timestamp()

    @property
    def _last_modified(self):
        """Get the date/time the file was last modified in units of seconds since epoch (posix time).

        :return float|None:
        """
        if self.exists_in_cloud:
            last_modified = self._cloud_metadata.get("updated")

            if last_modified is None:
                return None

            return last_modified.timestamp()

        try:
            return os.path.getmtime(self.local_path)
        except FileNotFoundError:
            return None

    @property
    def size_bytes(self):
        """Get the size of the datafile in bytes.

        :return float|None:
        """
        if self.exists_in_cloud:
            return self._cloud_metadata.get("size")

        try:
            return os.path.getsize(self.local_path)
        except FileNotFoundError:
            return None

    @property
    def exists_locally(self):
        """Return `True` if the file exists locally.

        :return bool:
        """
        return self._local_path is not None

    @property
    def local_path(self):
        """Get the local path for the datafile, downloading it from the cloud to a temporary file if necessary. If
        downloaded, the local path is added to a cache to avoid downloading again in the same runtime.

        :return str:
        """
        if self._local_path:
            return self._local_path

        return self.download()

    @property
    def path(self):
        """Alias to the `local_path` property.

        :return str:
        """
        return self.local_path

    @local_path.setter
    def local_path(self, path):
        """Set the local path of the datafile and:
        - If it exists in the cloud, download the contents of the corresponding cloud file to the new local path
        - If it only exists locally, copy the contents of the old local path to the new local path

        :param str path:
        :raise octue.exceptions.CloudLocationNotSpecified: if `path` is `None` and the datafile doesn't exist in the cloud
        :raise FileExistsError: if the new path corresponds to an existing local file
        :return None:
        """
        if path is None:
            if not self.exists_in_cloud:
                raise CloudLocationNotSpecified(
                    "The local path cannot be reset because this datafile only exists locally."
                )

            self._local_path = None
            return

        if os.path.exists(path):
            raise FileExistsError(
                "Only a path not corresponding to an existing file can be used. This is because the contents of the "
                "existing file would overwrite the existing file at the given path."
            )

        if self.exists_in_cloud:
            GoogleCloudStorageClient().download_to_file(local_path=path, cloud_path=self.cloud_path)
        else:
            os.makedirs(os.path.split(path)[0], exist_ok=True)
            shutil.copy(self._local_path, path)

        self._local_path = os.path.abspath(path)

    @property
    def open(self):
        """Get a context manager for handling the opening and closing of the datafile for reading/editing.

        :return type: the class octue.resources.datafile._DatafileContextManager
        """
        return functools.partial(_DatafileContextManager, self)

    @property
    def _local_metadata_path(self):
        """Get the path to the datafile's local metadata file (if the datafile exists locally).

        :return str|None:
        """
        if not self.exists_locally:
            return None

        return os.path.join(os.path.dirname(self._local_path), METADATA_FILENAME)

    def __enter__(self):
        self._open_context_manager = self.open(**self._open_attributes)
        return self, self._open_context_manager.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._open_context_manager.__exit__(exc_type, exc_val, exc_tb)
        del vars(self)["_open_context_manager"]

    def __lt__(self, other):
        if not isinstance(other, Datafile):
            raise TypeError(f"An object of type {type(self)} cannot be compared with {type(other)}.")
        return self.name < other.name

    def __gt__(self, other):
        if not isinstance(other, Datafile):
            raise TypeError(f"An object of type {type(self)} cannot be compared with {type(other)}.")
        return self.name > other.name

    def upload(self, cloud_path=None, update_cloud_metadata=True):
        """Upload a datafile to Google Cloud Storage.

        :param str|None cloud_path: full path to cloud storage location to store datafile at (e.g. `gs://bucket_name/path/to/file.csv`)
        :param bool update_cloud_metadata: if `True`, update the metadata of the datafile in the cloud at upload time
        :return str: gs:// path for datafile
        """
        cloud_path = self._get_cloud_location(cloud_path)

        self._get_cloud_metadata()

        storage_client = GoogleCloudStorageClient()

        # If the there is no cloud file or if the datafile's file has been changed locally, overwrite its cloud copy.
        if not storage_client.exists(cloud_path) or self.cloud_hash_value != self.hash_value:
            storage_client.upload_file(
                local_path=self.local_path,
                cloud_path=cloud_path,
                metadata=self.metadata(),
            )

        if update_cloud_metadata:
            # If the datafile's metadata has been changed locally, update the cloud file's metadata.
            local_metadata = self.metadata(use_octue_namespace=False)

            if self._cloud_metadata.get("custom_metadata") != local_metadata:
                self.update_cloud_metadata()

        return self.cloud_path

    def download(self, local_path=None):
        """Download the file from the cloud to the given local path or a random temporary path.

        :param str|None local_path:
        :raise CloudLocationNotSpecified: if the datafile does not exist in the cloud
        :return str: path to local file
        """
        if not self.exists_in_cloud:
            raise CloudLocationNotSpecified("Cannot download a file that doesn't exist in the cloud.")

        # Avoid downloading to a local path if the datafile has already been downloaded to it.
        if (local_path is None and self._local_path is not None) or (
            local_path is not None and local_path == self._local_path
        ):
            return self._local_path

        if local_path is not None:
            self._local_path = os.path.abspath(local_path)
        else:
            self._local_path = tempfile.NamedTemporaryFile(delete=False).name

        # Download from a URL.
        if storage.path.is_url(self.cloud_path):
            with open(self._local_path, "wb") as f:
                f.write(requests.get(self.cloud_path).content)

        # Download from a cloud URI.
        else:
            try:
                GoogleCloudStorageClient().download_to_file(local_path=self._local_path, cloud_path=self.cloud_path)

            except google.api_core.exceptions.NotFound as e:
                # If in reading mode, raise an error if no file exists at the path; if in a writing mode, create a new file.
                if self._open_attributes["mode"] == "r":
                    raise e

        # Now use hash value of local file instead of cloud file.
        self.reset_hash()
        return self._local_path

    def metadata(self, include_sdk_version=True, use_octue_namespace=True):
        """Get the datafile's metadata in a serialised form (i.e. the attributes `id`, `timestamp`, `labels`, `tags`,
        and `sdk_version`).

        :param bool include_sdk_version: if `True`, include the `octue` version that instantiated the datafile in the metadata
        :param bool use_octue_namespace: if `True`, prefix metadata names with "octue__"
        :return dict:
        """
        metadata = super().metadata(include_sdk_version=include_sdk_version)

        if not use_octue_namespace:
            return metadata

        return {f"{OCTUE_METADATA_NAMESPACE}__{key}": value for key, value in metadata.items()}

    def update_metadata(self):
        """If the datafile is cloud-based, update its cloud metadata; otherwise, update its local metadata.

        :return None:
        """
        if self.exists_in_cloud:
            self.update_cloud_metadata()
            return

        self.update_local_metadata()

    def update_cloud_metadata(self):
        """Update the cloud metadata for the datafile.

        :return None:
        """
        if not self.cloud_path:
            self._raise_cloud_location_error()

        GoogleCloudStorageClient().overwrite_custom_metadata(metadata=self.metadata(), cloud_path=self.cloud_path)

    def update_local_metadata(self):
        """Create or update the local octue metadata file with the datafile's metadata.

        :return None:
        """
        existing_metadata_records = load_local_metadata_file(self._local_metadata_path)

        if not existing_metadata_records.get("datafiles"):
            existing_metadata_records["datafiles"] = {}

        existing_metadata_records["datafiles"][self.name] = self.metadata(use_octue_namespace=False)
        overwrite_local_metadata_file(data=existing_metadata_records, path=self._local_metadata_path)

    def generate_signed_url(self, expiration=datetime.timedelta(days=7)):
        """Generate a signed URL for the datafile.

        :param datetime.datetime|datetime.timedelta expiration: the amount of time or date after which the URL should expire
        :return str: the signed URL for the datafile
        """
        if not self.exists_in_cloud:
            raise CloudLocationNotSpecified(
                f"{self!r} must exist in the cloud for a signed URL to be generated for it."
            )

        return GoogleCloudStorageClient().generate_signed_url(cloud_path=self.cloud_path, expiration=expiration)

    def _instantiate_from_cloud_object(self, path, local_path, ignore_stored_metadata):
        """Instantiate the datafile from a cloud object.

        :param str path: the cloud path to a cloud object
        :param str|None local_path: a local path to use for opening or modifying the cloud object locally
        :param bool ignore_stored_metadata: if `True`, don't use any metadata stored for this datafile in the cloud
        :return None:
        """
        self._cloud_path = path

        if not ignore_stored_metadata:
            self._use_cloud_metadata()

        if local_path:
            # If there is no file at the given local path or the file is different to the one in the cloud, download
            # the cloud file locally.
            if not os.path.exists(local_path) or self.cloud_hash_value != calculate_hash(local_path):
                self.download(local_path)
            else:
                self._local_path = local_path

    def _instantiate_from_local_path(self, path, cloud_path, ignore_stored_metadata):
        """Instantiate the datafile from a local path.

        :param str path: the path to a local file
        :param str|None cloud_path: a cloud path to upload the datafile to and mirror any changes made to it locally
        :param bool ignore_stored_metadata: if `True`, don't use any metadata stored for this datafile locally
        :return None:
        """
        self._local_path = os.path.abspath(path)

        if not ignore_stored_metadata:
            self._use_local_metadata()

        if cloud_path:
            self.cloud_path = cloud_path

    def _get_cloud_metadata(self):
        """Get the cloud metadata for the datafile and store it without updating the datafile's metadata.

        :return None:
        """
        if not self.cloud_path:
            self._raise_cloud_location_error()

        # Skip getting metadata for now if the cloud path is a signed URL.
        if storage.path.is_url(self.cloud_path):
            return

        cloud_metadata = GoogleCloudStorageClient().get_metadata(cloud_path=self.cloud_path)

        if cloud_metadata:
            cloud_metadata["custom_metadata"] = {
                key.replace(f"{OCTUE_METADATA_NAMESPACE}__", ""): value
                for key, value in cloud_metadata["custom_metadata"].items()
            }

            self._cloud_metadata = cloud_metadata

    def _use_cloud_metadata(self):
        """Update the datafile's metadata from the metadata of the cloud object located at its path. If no metadata is
        stored for the datafile, do nothing.

        :return None:
        """
        self._get_cloud_metadata()
        cloud_custom_metadata = self._cloud_metadata.get("custom_metadata", {})

        if not cloud_custom_metadata:
            return

        self.immutable_hash_value = self.cloud_hash_value or EMPTY_STRING_HASH_VALUE
        self._set_metadata(cloud_custom_metadata)

    def _use_local_metadata(self):
        """Update the datafile's metadata from the local metadata records file. If no metadata is stored for the
        datafile, do nothing.

        :return None:
        """
        existing_metadata_records = load_local_metadata_file(self._local_metadata_path)
        datafile_metadata = existing_metadata_records.get("datafiles", {}).get(self.name, {})

        if not datafile_metadata:
            return

        self._set_metadata(datafile_metadata)

    def _set_metadata(self, metadata):
        """Set the datafile's metadata.

        :param dict metadata:
        :return None:
        """
        for attribute in self._METADATA_ATTRIBUTES:
            if attribute not in metadata:
                continue

            if attribute == "id":
                self._set_id(metadata["id"])
                continue

            setattr(self, attribute, metadata[attribute])

    def _calculate_hash(self):
        """Get the hash of the datafile according to the first of the following methods that is applicable:

        1. The hash of the file at its local path
        2. If it doesn't have a local path, the hash of the file at its cloud path
        3. If it doesn't have either of these, use the empty string hash value

        :return str:
        """
        if self._local_path and os.path.exists(self._local_path):
            # Calculate the hash of the file itself and then pass it to `Hashable` to include the hashes of any
            # attributes named in `self._ATTRIBUTES_TO_HASH`.
            hash_value = calculate_hash(self._local_path)
            return super()._calculate_hash(hash_value)
        else:
            return self.cloud_hash_value or EMPTY_STRING_HASH_VALUE


class _DatafileContextManager:
    """A context manager for opening datafiles for reading and writing locally or from the cloud. Its usage is analogous
    to the builtin open context manager. If opening a local datafile in write mode, the manager will attempt to
    determine if the folder path exists and, if not, will create the folder structure required to write the file.

    Usage:
    ```
    my_datafile = Datafile(path='subfolder/subsubfolder/my_datafile.json)
    with my_datafile.open('w') as fp:
        fp.write("{}")
    ```

    This is equivalent to the standard python:
    ```
    my_datafile = Datafile(path='subfolder/subsubfolder/my_datafile.json)
    os.makedirs(os.path.split(my_datafile.local_path)[0], exist_ok=True)
    with open(my_datafile.local_path, 'w') as fp:
        fp.write("{}")
    ```

    :param octue.resources.datafile.Datafile datafile:
    :param str mode: open the datafile for reading/editing in this mode (the mode options are the same as for the builtin `open` function)
    :param bool update_metadata: if this is `True`, update the stored metadata of the datafile when the context is exited
    :return None:
    """

    MODIFICATION_MODES = {"w", "a", "x", "+", "U"}

    def __init__(self, datafile, mode="r", update_metadata=True, **kwargs):
        self.datafile = datafile
        self.mode = mode
        self._update_metadata = update_metadata
        self.kwargs = kwargs
        self._fp = None

    def __enter__(self):
        """Open the datafile, first downloading it from the cloud if necessary.

        :return io.TextIOWrapper:
        """
        if "w" in self.mode:
            if self.datafile.exists_in_cloud and storage.path.is_url(self.datafile.cloud_path):
                raise ReadOnlyResource(f"{self.datafile} is read-only. Change the open mode to 'r' to continue.")

            os.makedirs(os.path.split(self.datafile.local_path)[0], exist_ok=True)

        if self.datafile.extension == "hdf5":
            try:
                pkg_resources.get_distribution("h5py")
                self._fp = h5py.File(self.datafile.local_path, self.mode, **self.kwargs)
            except pkg_resources.DistributionNotFound:
                raise ImportError(
                    "To use datafiles with HDF5 files, please install octue with the 'hdf5' option i.e. "
                    "`pip install octue[hdf5]`."
                )
        else:
            self._fp = open(self.datafile.local_path, self.mode, **self.kwargs)

        return self._fp

    def __exit__(self, *args):
        """Close the datafile, updating the corresponding file in the cloud if necessary and its metadata if
        `self._update_cloud_metadata` is True.

        :return None:
        """
        if self._fp is not None:
            self._fp.close()

        if any(character in self.mode for character in self.MODIFICATION_MODES):

            if self.datafile.exists_in_cloud:
                self.datafile.upload(update_cloud_metadata=self._update_metadata)

            elif self._update_metadata:
                self.datafile.update_local_metadata()


def calculate_hash(path):
    """Calculate the hash of the file at the given path."""
    hash = Checksum()

    with open(path, "rb") as f:
        # Read and update hash value in blocks of 4K.
        for byte_block in iter(lambda: f.read(4096), b""):
            hash.update(byte_block)

    return hash
